/* gindexedlist - an indexed list, wanna be part of the glib.
 * Copyright (C) 2004  Kristian Benoit
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>

#include "gindexedlist.h"

/* Playlist structure */
struct g_indexed_list_St {
	GHashTable *hash_table;
	GList *head;
	GList *tail;
	GHashFunc hash_func;
	GEqualFunc equal_func;
	GDestroyNotify key_destroy_func;
	GDestroyNotify value_destroy_func;
};

/* Type of data in the GList */
struct g_indexed_node
{
	gpointer key;
	gpointer value;
};

/* struct used in g_indexed_list_sort_with_data to pass
 * the function and the user_data to the callback */
struct g_indexed_list_sort_st
{
	GICompareDataFunc func;
	gpointer user_data;
};

/* private functions */
static GList *g_list_insert_after (GList *head, GList *sibling, gpointer data);
static GIndexedNode *g_indexed_list_create_data (gpointer key, gpointer value);
static void g_indexed_list_remove_node (GIndexedList *indexed_list, GList *node);
static void g_indexed_list_steal_node (GIndexedList *indexed_list, GList *node);
static gboolean g_indexed_list_valid_sub_list (GIndexedList *indexed_list, gpointer start_key, gpointer end_key);

static void g_indexed_list_copy_cb (gpointer key, gpointer value, gpointer user_data);
static gint g_indexed_list_sort_cb (gconstpointer a, gconstpointer b, gpointer user_func);
static gint g_indexed_list_sort_with_data_cb (gconstpointer a, gconstpointer b, gpointer pt_st);
static gboolean g_indexed_list_concat_cb (gpointer key, gpointer value, gpointer user_data);


/**** Constructor ****/
GIndexedList * g_indexed_list_new ( GHashFunc hash_func, GEqualFunc key_equal_func)
{
  	return g_indexed_list_new_full (hash_func, key_equal_func, NULL, NULL);
}

GIndexedList *
g_indexed_list_new_full ( GHashFunc hash_func,
								  GEqualFunc key_equal_func,
								  GDestroyNotify key_destroy_func,
								  GDestroyNotify value_destroy_func )
{
	GIndexedList *ret;
	ret = g_try_malloc (sizeof(GIndexedList));

	g_return_val_if_fail (ret, NULL);

	ret->head = NULL;
	ret->tail = NULL;
	ret->hash_table = g_hash_table_new_full ( hash_func,
															key_equal_func,
															NULL,
															NULL );
	ret->hash_func = hash_func;
	ret->equal_func = key_equal_func;
	ret->key_destroy_func = key_destroy_func;
	ret->value_destroy_func = value_destroy_func;
	return ret;
}

GIndexedList *g_indexed_list_destroy (GIndexedList *indexed_list)
{
	g_indexed_list_clear (indexed_list);
	g_hash_table_destroy (indexed_list->hash_table);
	g_free (indexed_list);
	return NULL;
}

static void g_indexed_list_copy_cb (gpointer key, gpointer value, gpointer user_data)
{
	GIndexedList *indexed_list;

	indexed_list = (GIndexedList *)user_data;

	g_return_if_fail (indexed_list);

	g_indexed_list_append (indexed_list, key, value);
}

GIndexedList *g_indexed_list_copy (GIndexedList *indexed_list)
{
	g_return_val_if_fail (indexed_list, NULL);

	GIndexedList *new_indexed_list;
	new_indexed_list = g_indexed_list_new_full (indexed_list->hash_func,
															  indexed_list->equal_func,
															  indexed_list->key_destroy_func,
															  indexed_list->value_destroy_func);

	g_return_val_if_fail (new_indexed_list, NULL);

	g_indexed_list_foreach (indexed_list, g_indexed_list_copy_cb, (gpointer)new_indexed_list);
	return new_indexed_list;
}

static GIndexedNode *g_indexed_list_create_data (gpointer key, gpointer value)
{
	GIndexedNode *data;

	data = g_try_malloc (sizeof (GIndexedNode));
	
	g_return_val_if_fail (data, NULL);

	data->key = key;
	data->value = value;
	
	return data;
}

/* This function should go in the glist object. */
static GList *g_list_insert_after (GList *head, GList *sibling, gpointer data)
{
	GList *node;

	node = g_list_alloc();
	node->data = data;

	if (!(sibling && head))
	{
		node->prev = NULL;
		node->next = NULL;
		return node;
	}
	else if (!sibling)
	{
		sibling = g_list_last(head);
	}

	node->next = sibling->next;
	node->prev = sibling;
	sibling->next = node;
	if (node->next)
		node->next->prev = node;
	return head;
}

/**** insertion functions ****/
void g_indexed_list_append (GIndexedList *indexed_list, gpointer key, gpointer value)
{
	GList *node;
	gpointer data;

	g_assert (indexed_list);

	data = (gpointer)g_indexed_list_create_data (key, value);

	indexed_list->head = g_list_insert_after (indexed_list->head, indexed_list->tail, data);
	if (indexed_list->tail)
		node = g_list_next (indexed_list->tail);
	else
		node = indexed_list->head;
	g_hash_table_insert (indexed_list->hash_table, key, node);

	indexed_list->tail = node;
}

void g_indexed_list_prepend (GIndexedList *indexed_list, gpointer key, gpointer value)
{
	gpointer data;

	data = (gpointer)g_indexed_list_create_data (key, value);
	indexed_list->head = g_list_prepend (indexed_list->head, data);
	g_hash_table_insert (indexed_list->hash_table, key, indexed_list->head);

	if (!indexed_list->head->next)
		indexed_list->tail = indexed_list->head;
}

void g_indexed_list_insert (GIndexedList *indexed_list, gpointer key, gpointer value, guint position)
{
	GList *node, *last;
	gpointer data;

	g_assert (indexed_list);

	data = (gpointer)g_indexed_list_create_data (key, value);
	node->data = data;

	for (; position > 0 && node ; position--)
  	{
		node = g_list_next (node);
	}

	if (node)
	{
		indexed_list->head = g_list_insert_before (indexed_list->head, node, data);
		node = g_list_previous (node);
	}
  	else
	{
		indexed_list->head = g_list_insert_after (indexed_list->head, indexed_list->tail, data);
		indexed_list->tail = g_list_next (indexed_list->tail);
		node = indexed_list->tail ? indexed_list->tail : indexed_list->head;
	}

	g_hash_table_insert (indexed_list->hash_table, key, node);
}

void g_indexed_list_insert_before (GIndexedList *indexed_list, gpointer sibling_key, gpointer key, gpointer value)
{
	GList *node;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, sibling_key);

	if (node)
	{
		gpointer data;
		data = (gpointer)g_indexed_list_create_data (key, value);

		indexed_list->head = g_list_insert_before (indexed_list->head, node, data);
		node = g_list_previous (node);
		g_hash_table_insert (indexed_list->hash_table, key, node);
	}
	else
		g_indexed_list_prepend (indexed_list, key, value);
}

void g_indexed_list_insert_sorted (GIndexedList *indexed_list, gpointer key, gpointer value, GCompareFunc func)
{
	GList *node, *last;
	gpointer data;

	g_assert (indexed_list && func);

	data = (gpointer)g_indexed_list_create_data (key, value);
	
	for (node = indexed_list->head ; node && func(value, node->data) > 0 ; node = g_list_next (node));
	
	if (node)
	{
		indexed_list->head = g_list_insert_before (indexed_list->head, node, data);
		node = g_list_previous (node);
		g_hash_table_insert (indexed_list->hash_table, key, node);
	}
	else
		g_indexed_list_append (indexed_list, key, value);
}

void g_indexed_list_replace (GIndexedList *indexed_list, gpointer key, gpointer value)
{
	GList *node;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, key);

	if (!node)
	{
		g_indexed_list_append (indexed_list, key, value);
		return;
	}

	data = (GIndexedNode *)node->data;

	g_hash_table_remove (indexed_list->hash_table, key);

	if (indexed_list->key_destroy_func)
		indexed_list->key_destroy_func (data->key);
	if (indexed_list->value_destroy_func)
		indexed_list->value_destroy_func (data->value);

	data->key = key;
	data->value = value;
	g_hash_table_insert (indexed_list->hash_table, key, node);
}

/**** move functions ****/
static gboolean g_indexed_list_invalid_sub_list (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	GList *node, *end;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_val_if_fail (node && end, TRUE);

	while (node && node != end) node = g_list_next (node);
	
	return node ? FALSE : TRUE;
}

static gboolean g_indexed_list_invalid_sub_list_with_sibling (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, gpointer sibling_key)
{
	GList *node, *end, *sibling;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);
	sibling = g_hash_table_lookup (indexed_list->hash_table, sibling_key);

	g_return_val_if_fail (node && end && sibling, TRUE);

	while (node && node != sibling && node != end) node = g_list_next (node);
	
	return (node && node != sibling) ? FALSE : TRUE;
}

void g_indexed_list_move_first (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_first_nocheck (indexed_list, start_key, end_key);
}

void g_indexed_list_move_first_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	GList *start, *end;

	g_assert (indexed_list);

	start = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_if_fail (start && end);

	if (!start->prev)
		return;
	start->prev->next = end->next;
	if (end->next)
		end->next->prev = start->prev;
	else
		indexed_list->tail = start->prev;

	start->prev = NULL;
	end->next = indexed_list->head;
	indexed_list->head->prev = end;
	indexed_list->head = start;
}

void g_indexed_list_move_last (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_last_nocheck (indexed_list, start_key, end_key);
}

void g_indexed_list_move_last_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	GList *start, *end, *last;

	g_assert (indexed_list);

	start = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_if_fail (start && end);

	if (!end->next)
		return;

	last = g_list_last (end);
	end->next->prev = start->prev;
	if (start->prev)
		start->prev->next = end->next;
	else
		indexed_list->head = end->next;

	end->next = NULL;
	start->prev = last;
	last->next = start;

	indexed_list->tail = end;
}

void g_indexed_list_move_before (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, gpointer sibling_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list_with_sibling (indexed_list, start_key, end_key, sibling_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_before_nocheck (indexed_list, start_key, end_key, sibling_key);
}

void g_indexed_list_move_before_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, gpointer sibling_key)
{
	GList *start, *end, *sibling;

	g_assert (indexed_list);

	start = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);
	sibling = g_hash_table_lookup (indexed_list->hash_table, sibling_key);

	g_return_if_fail (start && end && sibling);

	if (sibling->prev == end)
		return;

	if (start->prev)
		start->prev->next = end->next;
	if (end->next)
		end->next->prev = start->prev;
	else
		indexed_list->tail = start->prev;

	start->prev = sibling->prev;
	end->next = sibling;
	sibling->prev = end;
	if (start->prev)
		start->prev->next = start;
	else
		indexed_list->head = start;
}

void g_indexed_list_move_after (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, gpointer sibling_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list_with_sibling (indexed_list, start_key, end_key, sibling_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_after_nocheck (indexed_list, start_key, end_key, sibling_key);
}

void g_indexed_list_move_after_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, gpointer sibling_key)
{
	GList *start, *end, *sibling;

	g_assert (indexed_list);

	start = g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = g_hash_table_lookup (indexed_list->hash_table, end_key);
	sibling = g_hash_table_lookup (indexed_list->hash_table, sibling_key);

	g_return_if_fail (start && end && sibling);

	if (sibling->next == start)
		return;

	if (start->prev)
		start->prev->next = end->next;
	else
		indexed_list->head = end->next;

	if (end->next)
		end->next->prev = start->prev;
	else
		indexed_list->tail = end;

	start->prev = sibling;
	end->next = sibling->next;
	sibling->next = start;
	if (end->next)
		end->next->prev = end;
}

void g_indexed_list_move_forward (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, guint position)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_forward_nocheck (indexed_list, start_key, end_key, position);
}

void g_indexed_list_move_forward_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, guint position)
{
	GList *node;
	GIndexedNode *data;

	if (position = 0)
		return;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_if_fail (node && g_list_next(node));

	for ( ; position && node ; position-- )
		node = g_list_next (node);

	if (node)
		data = (GIndexedNode *)node->data;
	else
		data = (GIndexedNode *)indexed_list->tail->data;

	g_indexed_list_move_after_nocheck (indexed_list, start_key, end_key, data->key);
}

void g_indexed_list_move_backward (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, guint position)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return;
	}
	g_indexed_list_move_backward_nocheck (indexed_list, start_key, end_key, position);
}

void g_indexed_list_move_backward_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key, guint position)
{
	GList *node;
	GIndexedNode *data;

	if (position = 0)
		return;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_if_fail (node && g_list_previous(node));

	for ( ; position && node ; position-- )
		node = g_list_previous (node);

	if (node)
		data = (GIndexedNode *)node->data;
	else
		data = (GIndexedNode *)indexed_list->head->data;

	g_indexed_list_move_before_nocheck (indexed_list, start_key, end_key, data->key);
}

/**** remove functions ****/
static void g_indexed_list_remove_node (GIndexedList *indexed_list, GList *node)
{
	GIndexedNode *data;
	gpointer key;

	g_assert (node);

	data = (GIndexedNode *)node->data;

	g_assert (data);
	
	key = data->key;

	if (node == indexed_list->tail)
		indexed_list->tail = g_list_previous (indexed_list->tail);

	g_hash_table_remove (indexed_list->hash_table, key);
	if (indexed_list->key_destroy_func)
		indexed_list->key_destroy_func (data->key);
	if (indexed_list->value_destroy_func)
		indexed_list->value_destroy_func (data->value);

	g_free (data);
	indexed_list->head = g_list_delete_link (indexed_list->head, node);
}

static void g_indexed_list_steal_node (GIndexedList *indexed_list, GList *node)
{
	GIndexedNode *data;
	gpointer key;

	g_assert (node);

	data = (GIndexedNode *)node->data;

	g_assert (data);
	
	key = data->key;

	if (node == indexed_list->tail)
		indexed_list->tail = g_list_previous (indexed_list->tail);

	g_hash_table_remove (indexed_list->hash_table, key);
	g_free (data);
	indexed_list->head = g_list_delete_link (indexed_list->head, node);
}

gboolean g_indexed_list_remove (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	gboolean ret;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);

	g_return_val_if_fail (node, FALSE);

	g_indexed_list_remove_node (indexed_list, node);

	return TRUE;
}

gboolean g_indexed_list_steal (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	gboolean ret;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, key);

	g_return_val_if_fail (node, FALSE);

	g_indexed_list_steal_node (indexed_list, node);

	return ret;
}

guint g_indexed_list_foreach_remove (GIndexedList *indexed_list, GHRFunc func, gpointer user_data)
{
	GList *node, *tmpnode;
	guint nb_removed;
	GIndexedNode *data;

	g_assert (indexed_list && func);

	nb_removed = 0;
	node = indexed_list->head;
	while (node)
	{
		data = (GIndexedNode *)node->data;
		if (func(data->key, data->value, user_data))
		{
			tmpnode = g_list_next (node);

			g_indexed_list_remove_node (indexed_list, node);

			node = tmpnode;
			nb_removed++;
		}
		else
			node = g_list_next (node);
	}
	return nb_removed;
}

guint g_indexed_list_foreach_steal (GIndexedList *indexed_list, GHRFunc func, gpointer user_data)
{
	GList *node, *tmpnode;
	guint nb_removed;
	GIndexedNode *data;

	g_assert (indexed_list);

	nb_removed = 0;
	node = indexed_list->head;
	while (node)
	{
		data = (GIndexedNode *)node->data;
		if (func(data->key, data->value, user_data))
		{
			tmpnode = g_list_next (node);

			g_indexed_list_steal_node (indexed_list, node);

			node = tmpnode;
			nb_removed++;
		}
		else
		  	node = g_list_next (node);
	}
	return nb_removed;
}

guint g_indexed_list_crop (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return 0;
	}
	return g_indexed_list_crop_nocheck (indexed_list, start_key, end_key);
}

guint g_indexed_list_crop_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	GList *node, *tmpnode, *start, *end;
	guint nb_removed;
	GIndexedNode *data;

	g_assert (indexed_list);

	start = (GList *)g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = (GList *)g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_val_if_fail (start && end, 0);

	nb_removed = 0;

	node = g_list_previous (start);
	while (node)
	{
		tmpnode = g_list_previous (node);

		g_indexed_list_remove_node (indexed_list, node);

		node = tmpnode;
		nb_removed++;
	}
	node = g_list_next (end);
	while (node)
	{
		tmpnode = g_list_next (node);

		g_indexed_list_remove_node (indexed_list, node);

		node = tmpnode;
		nb_removed++;
	}
	return nb_removed;
}

guint g_indexed_remove_sub (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	g_assert (indexed_list);
	if (g_indexed_list_invalid_sub_list (indexed_list, start_key, end_key))
	{
		g_warning ("Invalid sub list provided to %s", __PRETTY_FUNCTION__);
		return 0;
	}
	return g_indexed_list_remove_sub_nocheck (indexed_list, start_key, end_key);
}

guint g_indexed_list_remove_sub_nocheck (GIndexedList *indexed_list, gpointer start_key, gpointer end_key)
{
	GList *node, *tmpnode, *end;
	guint nb_removed;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, start_key);
	end = (GList *)g_hash_table_lookup (indexed_list->hash_table, end_key);

	g_return_val_if_fail (node && end, 0);

	nb_removed = 0;

	end = g_list_next (end);

	while (node != end)
	{
		tmpnode = g_list_next (node);

		g_indexed_list_remove_node (indexed_list, node);

		node = tmpnode;
		nb_removed++;
	}
	return nb_removed;
}

void g_indexed_list_clear (GIndexedList *indexed_list)
{
	GList *node, *tmpnode;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = indexed_list->head;
	while (node)
	{
		data = (GIndexedNode *)node->data;
		tmpnode = g_list_next (node);

		g_hash_table_remove (indexed_list->hash_table, data->key);
		if (indexed_list->key_destroy_func)
			indexed_list->key_destroy_func(data->key);
		if (indexed_list->value_destroy_func)
			indexed_list->value_destroy_func(data->value);
		g_free (data);
		g_list_delete_link (indexed_list->head, node);

		node = tmpnode;
	}
	indexed_list->head = NULL;
	indexed_list->tail = NULL;
}

/**** Info functions ****/
guint g_indexed_list_length (GIndexedList *indexed_list)
{
	g_assert (indexed_list);
	return g_hash_table_size (indexed_list->hash_table);
}

guint g_indexed_list_length_before (GIndexedList *indexed_list, gpointer key)
{
	GList *node, *iterator;
	guint ret;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, 0);

	ret = 0;
	for (iterator = indexed_list->head ; iterator && iterator != node ; iterator = g_list_next (iterator)) ret++;

	return ret;
}

guint g_indexed_list_length_after (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	guint ret;

	g_assert (indexed_list);

	node = g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, 0);

	ret = g_list_length (g_list_next (node));
	return ret;
}

/**** manipulation functions ****/
void g_indexed_list_reverse (GIndexedList *indexed_list)
{
	g_assert (indexed_list);
	indexed_list->head = g_list_reverse (indexed_list->head);
}

static gint g_indexed_list_sort_cb (gconstpointer a, gconstpointer b, gpointer user_func)
{
	GIndexedNode *data_a, *data_b;
	GICompareFunc func;

	data_a = (GIndexedNode *)a;
	data_b = (GIndexedNode *)b;
	func = (GICompareFunc)user_func;

	return func (data_a->key, data_a->value, data_b->key, data_b->value);
}
void g_indexed_list_sort (GIndexedList *indexed_list, GICompareFunc compare_func)
{
	g_assert (indexed_list);

	indexed_list->head = g_list_sort_with_data (indexed_list->head, g_indexed_list_sort_cb, (gpointer)compare_func);
}

static gint g_indexed_list_sort_with_data_cb (gconstpointer a, gconstpointer b, gpointer pt_st)
{
	GIndexedNode *data_a, *data_b;
	GICompareDataFunc func;
	gpointer user_data;
	struct g_indexed_list_sort_st *st;

	data_a = (GIndexedNode *)a;
	data_b = (GIndexedNode *)b;
	st = (struct g_indexed_list_sort_st *)pt_st;
	func = st->func;
	user_data = st->user_data;

	return func (data_a->key, data_a->value, data_b->key, data_b->value, user_data);
}
void g_indexed_list_sort_with_data (GIndexedList *indexed_list, GICompareDataFunc compare_func, gpointer user_data)
{
	struct g_indexed_list_sort_st st = {compare_func, user_data};

	g_assert (indexed_list);

	indexed_list->head = g_list_sort_with_data (indexed_list->head, g_indexed_list_sort_with_data_cb, (gpointer)&st);
}

void g_indexed_list_shuffle (GIndexedList *indexed_list)
{
	gint len, i, j;
	GList *node, **ptrs;

	g_assert (indexed_list);

	len = g_indexed_list_length (indexed_list);

	if (len < 2)
		return;

	ptrs = g_new (GList *, len);

	for ( node = indexed_list->head, i = 0;
			i < len && node;
			node = g_list_next (node), i++ )
		ptrs[i] = node;

	j = random() % len;
	indexed_list->head = ptrs[j];
	indexed_list->tail = ptrs[j];
	ptrs[j]->next = NULL;
	ptrs[j] = ptrs[0];

	for (i = 1; i < len; i++) {
		j = random() % (len - i);
		indexed_list->head->prev = ptrs[i + j];
		ptrs[i + j]->next = indexed_list->head;
		indexed_list->head = ptrs[i + j];
		ptrs[i + j] = ptrs[i];
	}
	indexed_list->head->prev = NULL;

	g_free (ptrs);
}

static gboolean g_indexed_list_concat_cb (gpointer key, gpointer value, gpointer user_data)
{
	GIndexedList *indexed_list;
	indexed_list = (GIndexedList *)user_data;

	g_assert (indexed_list);

	g_hash_table_insert (indexed_list->hash_table, key, value);
	return TRUE;
}
void g_indexed_list_concat (GIndexedList *indexed_list1, GIndexedList *indexed_list2)
{
	g_assert (indexed_list1);

	g_hash_table_foreach_remove (indexed_list2->hash_table, g_indexed_list_concat_cb, indexed_list1);
	indexed_list1->head = g_list_concat(indexed_list1->head, indexed_list2->head);
	indexed_list2->head = NULL;
}

void g_indexed_list_foreach (GIndexedList *indexed_list, GHFunc func, gpointer user_data)
{
	GList *node;
	GIndexedNode *data;
	g_assert (indexed_list && func);

	for (node = indexed_list->head ; node ; node = g_list_next (node))
	{
		data = (GIndexedNode *)node->data;
		func (data->key, data->value, user_data);
	}
}

/**** assessor functions ****/
gpointer g_indexed_list_lookup (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, NULL);

	data = (GIndexedNode *)node->data;

	return data->value;
}

gboolean g_indexed_list_lookup_extended (GIndexedList *indexed_list, gconstpointer lookup_key, gpointer *orig_key, gpointer *value)
{
	GList *node;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, lookup_key);
	if (! node)
		return FALSE;

	data = (GIndexedNode *)node->data;

	*orig_key = data->key;
	*value = data->value;
	return TRUE;
}

gpointer g_indexed_list_find (GIndexedList *indexed_list, GHRFunc predicate, gpointer user_data)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list && predicate);

	for (node = indexed_list->head ; node ; node = g_list_next (node))
	{
		data = (GIndexedNode *)node->data;
		if (predicate (data->key, data->value, user_data))
			break;
	}

	if (node)
		ret = data->value;
	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_first (GIndexedList *indexed_list)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = indexed_list->head;

	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_last (GIndexedList *indexed_list)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = indexed_list->tail;

	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_previous (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, NULL);

	node = g_list_previous (node);

	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_next (GIndexedList *indexed_list, gpointer key)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, NULL);

	node = g_list_next (node);

	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_nth (GIndexedList *indexed_list, gpointer key, guint n)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, NULL);

	node = g_list_nth (node, n);

	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}

gpointer g_indexed_list_nth_prev (GIndexedList *indexed_list, gpointer key, guint n)
{
	GList *node;
	gpointer *ret;
	GIndexedNode *data;

	g_assert (indexed_list);

	node = (GList *)g_hash_table_lookup (indexed_list->hash_table, key);
	g_return_val_if_fail (node, NULL);

	node = g_list_nth_prev (node, n);
	
	if (node)
	{
		data = (GIndexedNode *)node->data;
		ret = data->value;
	}
  	else
		ret = NULL;

	return ret;
}
